import logging

import torch

logger = logging.getLogger(__name__)


def pytorch_cos_sim(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.

    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    return cos_sim(a, b)


def cos_sim(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.

    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    a_norm = torch.nn.functional.normalize(a, p=2, dim=1)
    b_norm = torch.nn.functional.normalize(b, p=2, dim=1)
    return torch.mm(a_norm, b_norm.transpose(0, 1))


#
#
# def dot_score(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
#     """
#     Computes the dot-product dot_prod(a[i], b[j]) for all i and j.
#
#     :return: Matrix with res[i][j]  = dot_prod(a[i], b[j])
#     """
#     if not isinstance(a, torch.Tensor):
#         a = torch.tensor(a)
#
#     if not isinstance(b, torch.Tensor):
#         b = torch.tensor(b)
#
#     if len(a.shape) == 1:
#         a = a.unsqueeze(0)
#
#     if len(b.shape) == 1:
#         b = b.unsqueeze(0)
#
#     return torch.mm(a, b.transpose(0, 1))
#
#
# def pairwise_dot_score(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
#     """
#     Computes the pairwise dot-product dot_prod(a[i], b[i])
#
#     :return: Vector with res[i] = dot_prod(a[i], b[i])
#     """
#     if not isinstance(a, torch.Tensor):
#         a = torch.tensor(a)
#
#     if not isinstance(b, torch.Tensor):
#         b = torch.tensor(b)
#
#     return (a * b).sum(dim=-1)
#
#
# def pairwise_cos_sim(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
#     """
#     Computes the pairwise cossim cos_sim(a[i], b[i])
#
#     :return: Vector with res[i] = cos_sim(a[i], b[i])
#     """
#     if not isinstance(a, torch.Tensor):
#         a = torch.tensor(a)
#
#     if not isinstance(b, torch.Tensor):
#         b = torch.tensor(b)
#
#     return pairwise_dot_score(normalize_embeddings(a), normalize_embeddings(b))
#
#
# def pairwise_angle_sim(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
#     """
#     Computes the absolute normalized angle distance;
#     see AnglELoss or https://arxiv.org/abs/2309.12871v1
#     for more information.
#
#     :return: Vector with res[i] = angle_sim(a[i], b[i])
#     """
#
#     if not isinstance(x, torch.Tensor):
#         x = torch.tensor(x)
#
#     if not isinstance(y, torch.Tensor):
#         y = torch.tensor(y)
#
#     # modified from https://github.com/SeanLee97/AnglE/blob/main/angle_emb/angle.py
#     # chunk both tensors to obtain complex components
#     a, b = torch.chunk(x, 2, dim=1)
#     c, d = torch.chunk(y, 2, dim=1)
#
#     z = torch.sum(c**2 + d**2, dim=1, keepdim=True)
#     re = (a * c + b * d) / z
#     im = (b * c - a * d) / z
#
#     dz = torch.sum(a**2 + b**2, dim=1, keepdim=True) ** 0.5
#     dw = torch.sum(c**2 + d**2, dim=1, keepdim=True) ** 0.5
#     re /= dz / dw
#     im /= dz / dw
#
#     norm_angle = torch.sum(torch.concat((re, im), dim=1), dim=1)
#     return torch.abs(norm_angle)
#
#
# def normalize_embeddings(embeddings: torch.Tensor) -> torch.Tensor:
#     """
#     Normalizes the embeddings matrix, so that each sentence embedding has unit length
#     """
#     return torch.nn.functional.normalize(embeddings, p=2, dim=1)



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from transformers import BatchEncoding, DataCollatorWithPadding, PreTrainedTokenizer


class PairCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_key: str = 'query',
        positive_key: str = 'positive',
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        document_max_length: Optional[int] = None,
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_key = query_key
        self.positive_key = positive_key

        self.query_max_length: int
        self.document_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.document_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.document_max_length = tokenizer.model_max_length

        if document_max_length:
            self.document_max_length = document_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert len(features) > 0
        assert (
            self.query_key in features[0] and self.positive_key in features[0]
        ), f"PairCollator should have {self.query_key} and {self.positive_key} in features, while get {features[0]}"
        "you can set the custom key of query_key, positive_key during class init"

        query_texts = [feature["query"] for feature in features]
        pos_texts = [feature["positive"] for feature in features]

        query_inputs = self.tokenizer(
            query_texts,
            padding=True,
            max_length=self.query_max_length,
            truncation=True,
            return_tensors="pt",
        )
        pos_inputs = self.tokenizer(
            pos_texts,
            padding=True,
            max_length=self.document_max_length,
            truncation=True,
            return_tensors="pt",
        )

        return {"query": query_inputs, "positive": pos_inputs}


class TripletCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_key: str = 'query',
        positive_key: str = 'positive',
        negative_key: Optional[str] = 'negative',
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        document_max_length: Optional[int] = None,
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key

        self.query_max_length: int
        self.document_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.document_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.document_max_length = tokenizer.model_max_length

        if document_max_length:
            self.document_max_length = document_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert len(features) > 0
        assert (
            self.positive_key in features[0] and self.positive_key in features[0] and self.negative_key in features[0]
        ), f"TripletCollator should have {self.query_key}, {self.positive_key} and {self.negative_key} in features dict"
        "you can set the custom key of query_key, positive_key and negative_key during class init"

        query_texts = [feature[self.query_key] for feature in features]
        pos_texts = [feature[self.positive_key] for feature in features]
        neg_texts = [feature[self.negative_key] for feature in features]

        # if isinstance(query[0], list):
        #     query = sum(query, [])
        # if isinstance(document[0], list):
        #     document = sum(document, [])

        query_inputs = self.tokenizer(
            query_texts,
            padding=True,
            max_length=self.query_max_length,
            truncation=True,
            return_tensors="pt",
        )
        pos_inputs = self.tokenizer(
            pos_texts,
            padding=True,
            max_length=self.document_max_length,
            truncation=True,
            return_tensors="pt",
        )  # ["input_ids"]
        neg_inputs = self.tokenizer(
            neg_texts,
            padding=True,
            max_length=self.document_max_length,
            truncation=True,
            return_tensors="pt",
        )  # ["input_ids"]

        return {
            "query": query_inputs,
            "positive": pos_inputs,
            "negative": neg_inputs,
        }


class RerankCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_key: str = 'query',
        document_key: str = 'document',
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        document_max_length: Optional[int] = None,
    ):
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_key = query_key
        self.document_key = document_key

        self.query_max_length: int
        self.document_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        if document_max_length:
            self.document_max_length = document_max_length
        elif max_length:
            self.query_max_length = max_length
            self.document_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.document_max_length = tokenizer.model_max_length

    def __call__(self, features: Union[List[Dict[str, Any]], List]) -> BatchEncoding:
        assert len(features) > 0
        if isinstance(features[0], dict):
            assert (
                self.query_key in features[0] and self.document_key in features[0]
            ), f"RerankCollator should have {self.query_key} and {self.document_key} keys in features, "
            "and 'labels' during training"
            query_texts = [feature[self.query_key] for feature in features]
            document_texts = [feature[self.document_key] for feature in features]
        else:
            query_texts = [feature[0] for feature in features]
            document_texts = [feature[1] for feature in features]

        batch = self.tokenizer(
            text=query_texts,
            text_pair=document_texts,
            padding=True,
            truncation=True,
            max_length=self.query_max_length,
            return_tensors="pt",
        )

        if 'labels' in features[0].keys():
            labels = [feature['labels'] for feature in features]
            batch['labels'] = torch.tensor(labels, dtype=torch.float32)
        return batch


class ColBertCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_key: str = 'query',
        positive_key: str = 'positive',
        negative_key: str = 'negative',
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        document_max_length: Optional[int] = None,
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key

        self.query_max_length: int
        self.document_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.document_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.document_max_length = tokenizer.model_max_length

        if document_max_length:
            self.document_max_length = document_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert len(features) > 0
        assert (
            self.query_key in features[0] and self.positive_key in features[0]
        ), f"PairCollator should have {self.query_key} and {self.positive_key} in features, while get {features[0]}"
        "you can set the custom key of query_key, positive_key during class init"

        query_texts = [feature[self.query_key] for feature in features]
        pos_texts = [feature[self.positive_key] for feature in features]

        query_inputs = self.tokenizer(
            query_texts,
            padding=True,
            max_length=self.query_max_length,
            truncation=True,
            return_tensors="pt",
        )
        pos_inputs = self.tokenizer(
            pos_texts,
            padding=True,
            max_length=self.document_max_length,
            truncation=True,
            return_tensors="pt",
        )

        batch = {
            'query_input_ids': query_inputs['input_ids'],
            'query_attention_mask': query_inputs['attention_mask'],
            'pos_input_ids': pos_inputs['input_ids'],
            'pos_attention_mask': pos_inputs['attention_mask'],
        }
        if 'labels' in features[0].keys():
            labels = [feature['labels'] for feature in features]
            batch['labels'] = torch.tensor(labels, dtype=torch.float32)
        return batch



import torch
from torch import nn


class MoCoLoss(nn.Module):
    def __init__(self, temperature=0.05):
        super(MoCoLoss, self).__init__()
        self.temperature = temperature

    def forward(self, q, k):
        batch_size = q.shape[0]
        q = nn.functional.normalize(q, dim=1)  # Normalize query embeddings
        k = nn.functional.normalize(k, dim=1)  # Normalize key embeddings

        # Positive logits: Nx1
        logits_pos = torch.einsum('nc,nc->n', [q, k]).unsqueeze(-1)
        # Negative logits: NxK
        logits_neg = torch.einsum('nc,ck->nk', [q, self.queue.clone().detach()])

        logits = torch.cat([logits_pos, logits_neg], dim=1)

        logits /= self.temperature

        labels = torch.zeros(batch_size, dtype=torch.long).cuda()
        loss = nn.CrossEntropyLoss().cuda()(logits, labels)

        return loss



import logging
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch.nn as nn
from transformers import Trainer
from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.trainer_callback import TrainerCallback
from transformers.trainer_utils import EvalPrediction

from ..losses import InfoNCE, SimCSE, TripletLoss

logger = logging.getLogger(__name__)


class RetrievalTrainer(Trainer):
    def __init__(self, loss_fn, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = loss_fn

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        query = inputs["query"]
        pos = inputs["positive"]
        query_embeddings = model(query)
        pos_embeddings = model(pos)
        if 'negative' in inputs:
            neg = inputs["negative"]
            neg_embeddings = model(neg)
            loss = TripletLoss()(query_embeddings, pos_embeddings, neg_embeddings)
        else:
            loss = self.loss_fn(query_embeddings, pos_embeddings)
        if not return_outputs:
            return loss
        outputs = dict()
        outputs['query'] = query_embeddings
        outputs['positive'] = pos_embeddings
        if 'negative' in inputs:
            outputs['negative'] = neg_embeddings
        return (loss, outputs)

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)

        self.model.model.save_pretrained(output_dir)
        self.model.tokenizer.save_pretrained(output_dir)


class RerankTrainer(Trainer):
    def __init__(self, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        if not loss_fn:
            loss_fn = nn.BCEWithLogitsLoss()
        self.loss_fn = loss_fn

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        outputs: dict = model(**inputs)
        loss = outputs['loss']
        return (loss, outputs) if return_outputs else loss

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)

        self.model.model.save_pretrained(output_dir)
        self.model.tokenizer.save_pretrained(output_dir)



"""
`Representation Learning with Contrastive Predictive Coding
<https://arxiv.org/abs/1807.03748v2>`_
"""

import logging
from typing import Callable, Literal, Optional, Union

import torch
import torch.distributed.nn
import torch.nn as nn
import torch.nn.functional as F

logger = logging.getLogger(__name__)


class InfoNCE(nn.Module):
    """
    https://github.com/RElbers/info-nce-pytorch
    """

    def __init__(
        self,
        criterion: Union[nn.Module, Callable, None] = nn.CrossEntropyLoss(label_smoothing=0.05),
        temperature: float = 0.05,
        negative_mode: Literal['paired', 'unpaired'] = "unpaired",
    ):
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature
        self.negative_mode = negative_mode

    def forward(
        self,
        query_embeddings: torch.Tensor,
        positive_embeddings: torch.Tensor,
        negative_embeddings: Optional[torch.Tensor] = None,
    ):
        query_embeddings = F.normalize(query_embeddings, dim=-1)
        positive_embeddings = F.normalize(positive_embeddings, dim=-1)
        device = query_embeddings.device
        if negative_embeddings is None:
            logits1 = query_embeddings @ positive_embeddings.transpose(-2, -1)
            logits2 = logits1.T
            labels = torch.arange(len(logits1), dtype=torch.long, device=device)
            loss = (
                self.criterion(logits1 / self.temperature, labels) + self.criterion(logits2 / self.temperature, labels)
            ) / 2
            return loss
        else:
            negative_embeddings = F.normalize(negative_embeddings, dim=-1)
            positive_logit = torch.sum(query_embeddings * positive_embeddings, dim=1, keepdim=True)

            if self.negative_mode == 'unpaired':
                # Cosine between all query-negative combinations
                negative_logits = query_embeddings @ negative_embeddings.transpose(-2, -1)

            elif self.negative_mode == 'paired':
                query = query_embeddings.unsqueeze(1)
                negative_logits = query @ negative_embeddings.transpose(-2, -1)
                negative_logits = negative_logits.squeeze(1)

            else:
                raise ValueError(f"negative mode could chose 'unpaired' or 'paired', while got {self.negative_mode}")

            # First index in last dimension are the positive samples
            logits = torch.cat([positive_logit, negative_logits], dim=1)
            labels = torch.zeros(len(logits), dtype=torch.long, device=query_embeddings.device)
            return self.criterion(logits / self.temperature, labels)



import logging
import os
import time
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torch.distributed as dist
import torch.nn as nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from tqdm.autonotebook import trange
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoTokenizer,
    BatchEncoding,
    GenerationConfig,
    PreTrainedTokenizer,
)

from .pooling import AutoPooling
from .retrieval_auto import AutoModelForRetrieval
from .utils import batch_to_device, check_casual_lm, get_device_name

logger = logging.getLogger(__name__)


class AutoModelForEmbedding(nn.Module):
    """
    Loads or creates an Embedding model that can be used to map sentences / text.

    :param model_name_or_path: If it is a filepath on disc, it loads the model from that path. If it is not a path,
        it first tries to download a pre-trained SentenceTransformer model. If that fails, tries to construct a model
        from the Hugging Face Hub with that name.
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        max_length: Optional[int] = None,
        loss_fn: Optional[Callable] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        generation_args: Dict = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.pooling_method = pooling_method
        self.pooling = AutoPooling(pooling_method) if pooling_method else None
        self.loss_fn = loss_fn

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)
        else:
            logger.info('max_length will only work if the encode or forward function input text directly')

        self.max_length = max_length
        self.normalize_embeddings = normalize_embeddings

        self.query_instruction = query_instruction
        self.document_instruction = document_instruction
        if generation_args is not None:
            generation_config = self.model.generation_config.to_dict()
            generation_config.update(generation_args)
            generation_config.update({"pad_token_id": self.tokenizer.pad_token_id})
            self.model.generation_config = GenerationConfig(**generation_config)

        if device is None:
            self.device = get_device_name()
        else:
            self.device = device
        self.model.to(self.device)

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = False,
    ):
        if isinstance(inputs, (dict, BatchEncoding)):
            embeddings = self.forward_from_loader(inputs)
        elif isinstance(inputs, str) or (isinstance(inputs, list) and isinstance(inputs[0], str)):
            embeddings = self.forward_from_text(inputs)
        else:
            raise ValueError

        if labels is None or self.loss_fn is None:
            if return_dict:
                return {"sentence_embedding": embeddings}
            return embeddings
        else:
            outputs = dict()
            loss_output = self.loss_fn(embeddings, labels)
            outputs["loss"] = loss_output["loss"]
            outputs["sentence_embedding"] = loss_output["sentence_embedding"]
            return outputs

    def forward_from_loader(self, inputs, without_pooling: bool = False):
        model_output = self.model(inputs['input_ids'], inputs['attention_mask'])
        if self.pooling is not None and not without_pooling:
            if 'last_hidden_state' in model_output:
                last_hidden_state = model_output['last_hidden_state']
            elif 'hidden_states' not in model_output:
                last_hidden_state = model_output[0]
            else:
                hidden_states = model_output['hidden_states']
                # last_hidden_state = torch.cat([hidden_states[0], hidden_states[-1]])
                last_hidden_state = (hidden_states[0] + hidden_states[-1]) / 2.0
            embeddings = self.pooling(last_hidden_state, inputs["attention_mask"])

            if self.normalize_embeddings:
                embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

            return embeddings
        return model_output

    def forward_from_text(self, texts):
        batch_dict = self.tokenizer(
            texts,
            max_length=self.max_length,
            return_attention_mask=False,
            padding=False,
            truncation=True,
        )
        batch_dict["input_ids"] = [input_ids + [self.tokenizer.eos_token_id] for input_ids in batch_dict["input_ids"]]
        batch_dict = self.tokenizer.pad(batch_dict, padding=True, return_attention_mask=True, return_tensors="pt")
        batch_dict.pop("token_type_ids")
        return self.forward_from_loader(batch_dict)

    def encode(
        self,
        inputs: Union[DataLoader, Dict, List, str],
        batch_size: int = 128,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ):
        if isinstance(inputs, (DataLoader, BatchEncoding, Dict)):
            return self.encode_from_loader(
                loader=inputs,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        elif isinstance(inputs, (str, List, Tuple, pd.Series, np.ndarray)):
            return self.encode_from_text(
                sentences=inputs,
                batch_size=batch_size,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        else:
            raise ValueError(f'Input type: {type(inputs)}')

    def encode_from_loader(
        self,
        loader: DataLoader,
        convert_to_numpy: bool = True,
        device: str = None,
        normalize_embeddings: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ) -> Union[List[torch.Tensor], np.ndarray, torch.Tensor]:
        self.model.eval()
        self.model.to(device or self.device)

        all_embeddings = []
        with torch.no_grad():
            for idx, inputs in enumerate(tqdm(loader, desc="Encoding", disable=not show_progress_bar)):
                inputs_on_device = {k: v.to(self.device) for k, v in inputs.items()}
                embeddings = self.forward_from_loader(inputs_on_device)
                embeddings = embeddings.detach()
                if normalize_embeddings:
                    embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                all_embeddings.append(embeddings)
        if convert_to_numpy:
            all_embeddings = np.concatenate([emb.cpu().numpy() for emb in all_embeddings], axis=0)
        else:
            all_embeddings = torch.concat(all_embeddings)
        return all_embeddings

    def encode_from_text(
        self,
        sentences: Union[str, List[str], Tuple[str], pd.Series, np.ndarray],
        batch_size: int = 128,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ) -> Union[List[torch.Tensor], np.ndarray, torch.Tensor]:
        """
        Computes sentence embeddings from sentence-transformers library.

        :param sentences: the sentences to embed.
        :param batch_size: the batch size used for the computation.
        :param show_progress_bar: Whether to output a progress bar when encode sentences.
        :param output_value: The type of embeddings to return: "sentence_embedding" to get sentence embeddings,
            "token_embeddings" to get wordpiece token embeddings, and `None`, to get all output values. Defaults
            to "sentence_embedding".
        :param convert_to_numpy: Whether the output should be a list of numpy vectors. If False, it is a list of tensors
        :param convert_to_tensor: Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
        :param device: Which `torch.device` to use for the computation.
        :param normalize_embeddings: Whether to normalize returned vectors to have length 1. In that case,
            the faster dot-product (util.dot_score) instead of cosine similarity can be used.

        :return: By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned.
            If convert_to_numpy, a numpy matrix is returned.
        """
        self.model.eval()
        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG
            )

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value != "sentence_embedding":
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, "__len__"):
            sentences = [sentences]
            input_was_string = True

        if device is None:
            device = self.device

        self.to(device)

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sentence) for sentence in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenizer(
                sentences_batch,
                padding=True,
                truncation="longest_first",
                return_tensors="pt",
                max_length=self.max_length,
            )
            features = batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(features, return_dict=True)

                if output_value == "token_embeddings":
                    embeddings = []
                    for token_emb, attention in zip(out_features[output_value], out_features["attention_mask"]):
                        last_mask_id = len(attention) - 1
                        while last_mask_id > 0 and attention[last_mask_id].item() == 0:
                            last_mask_id -= 1

                        embeddings.append(token_emb[0 : last_mask_id + 1])
                elif output_value is None:  # Return all outputs
                    embeddings = []
                    for sent_idx in range(len(out_features["sentence_embedding"])):
                        row = {name: out_features[name][sent_idx] for name in out_features}
                        embeddings.append(row)
                else:  # Sentence embeddings
                    embeddings = out_features[output_value]
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

                    # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()

                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            if len(all_embeddings):
                all_embeddings = torch.stack(all_embeddings)
            else:
                all_embeddings = torch.Tensor()
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings

    def build_index(
        self,
        inputs: Union[DataLoader, Dict, List, str],
        index_path: Optional[str] = None,
        max_document_length: int = 512,
        split_documents: bool = False,
        batch_size: int = 128,
        show_progress_bar: bool = None,
        use_gpu: bool = False,
    ):
        import faiss

        logger.info("Start to build index")
        start_time = time.time()

        embeddings = self.encode(
            inputs, batch_size=batch_size, convert_to_numpy=True, show_progress_bar=show_progress_bar
        )
        embeddings = np.asarray(embeddings, dtype=np.float32)

        index = faiss.IndexFlatL2(len(embeddings[0]))
        if use_gpu and self.device == 'cuda':
            logger.info('Build index use faiss-gpu')
            co = faiss.GpuMultipleClonerOptions()
            co.shard = True
            co.useFloat16 = True
            index = faiss.index_cpu_to_all_gpus(index, co=co)
        index.add(embeddings)

        if index_path:
            logger.info(f'Save faiss index to: {index_path}')
            if os.path.isdir(index_path):
                index_path = index_path + 'faiss.index'
            if not os.path.exists(os.path.dirname(index_path)):
                os.makedirs(os.path.dirname(index_path))

            if use_gpu and self.device == 'cuda':
                index = faiss.index_gpu_to_cpu(index)
            faiss.write_index(index, index_path)

        logger.info(f'Build index successfully, saved in {index_path}, elapsed: {time.time() - start_time:.2}s')
        return index

    def add_to_index(self):
        return

    def set_train_type(self, train_type: Literal['pointwise', 'pairwise', 'listwise'], **kwargs):
        model_class = {'pointwise': self, 'pairwise': PairwiseModel, 'listwise': ListwiseModel}
        model_class = model_class.get(train_type.lower())
        return model_class(
            model=self.model,
            tokenizer=self.tokenizer,
            pooling_method=self.pooling_method,
            normalize_embeddings=self.normalize_embeddings,
            loss_fn=self.loss_fn,
            query_instruction=self.query_instruction,
            document_instruction=self.document_instruction,
            device=self.device,
            **kwargs,
        )

    @classmethod
    def as_retriever(cls, retrieval_args, **kwargs):
        embedding_model = cls(**kwargs)
        return AutoModelForRetrieval(embedding_model, **retrieval_args)

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: Optional[str] = None,
        pooling_method: Optional[str] = "cls",
        pretrained: bool = True,
        config_path: Optional[str] = None,
        trust_remote_code: bool = True,
        causal_lm: bool = False,
        custom_config_dict: Optional[Dict] = None,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        device: Optional[str] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        **kwargs,
    ):
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        if config_path:
            config = AutoConfig.from_pretrained(
                config_path, output_hidden_states=True, trust_remote_code=trust_remote_code
            )
        else:
            config = AutoConfig.from_pretrained(
                model_name_or_path, output_hidden_states=True, trust_remote_code=trust_remote_code
            )
        if custom_config_dict:
            config.update(custom_config_dict)

        if causal_lm or check_casual_lm(model_name_or_path):
            if pretrained:
                model = AutoModelForCausalLM.from_pretrained(
                    model_name_or_path, config=config, trust_remote_code=trust_remote_code, **kwargs
                )
            else:
                model = AutoModelForCausalLM.from_config(config)
        else:
            if pretrained:
                model = AutoModel.from_pretrained(
                    model_name_or_path, config=config, trust_remote_code=trust_remote_code, **kwargs
                )
            else:
                model = AutoModel.from_config(config)

        if use_fp16:
            model.half()

        if use_lora:
            # peft config and wrapping
            from peft import LoraConfig, TaskType, get_peft_model

            if not lora_config:
                raise ValueError("If use_lora is true, please provide a valid lora_config")
            model = get_peft_model(model, lora_config)
            model.print_trainable_parameters()

        return cls(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            device=device,
            query_instruction=query_instruction,
            document_instruction=document_instruction,
            **kwargs,
        )

    def save_pretrained(self, path: str, safe_serialization: bool = True):
        """
        Saves all model and tokenizer to path
        """
        logger.info("Save model to {}".format(path))
        state_dict = self.model.state_dict()
        state_dict = type(state_dict)({k: v.clone().cpu() for k, v in state_dict.items()})
        self.model.save_pretrained(path, state_dict=state_dict, safe_serialization=safe_serialization)
        self.tokenizer.save_pretrained(path)

    def push_to_hub(self, hub_model_id: str, private: bool = True, **kwargs):
        """push model to hub

        :param hub_model_id: str, hub model id.
        :param private: bool, whether push to private repo. Default True.
        :param kwargs: other kwargs for `push_to_hub` method.
        """
        self.tokenizer.push_to_hub(hub_model_id, private=private, **kwargs)
        self.backbone.push_to_hub(hub_model_id, private=private, **kwargs)

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """
        Help function to get the length for the input text. Text can be either
        a list of ints (which means a single text as input), or a tuple of list of ints
        (representing several text inputs to the model).
        """

        if isinstance(text, dict):  # {key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, "__len__"):  # Object has no len() method
            return 1
        elif len(text) == 0 or isinstance(text[0], int):  # Empty string or list of ints
            return len(text)
        else:
            return sum([len(t) for t in text])  # Sum of length of individual strings

    def _dist_gather_tensor(self, t: Optional[torch.Tensor]):
        if t is None:
            return None
        t = t.contiguous()

        all_tensors = [torch.empty_like(t) for _ in range(self.world_size)]
        dist.all_gather(all_tensors, t)

        all_tensors[self.process_rank] = t
        all_tensors = torch.cat(all_tensors, dim=0)

        return all_tensors


class PairwiseModel(AutoModelForEmbedding):
    """Pairwise Model wrapper
    - bi_encoder
    - cross_encoder
    - poly_encoder
    support: query + pos pair, or query + pos + neg triplet
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        loss_fn: Optional[Callable] = None,
        cross_encoder: bool = False,
        poly_encoder: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            normalize_embeddings=normalize_embeddings,
            loss_fn=None,
            **kwargs,
        )
        if loss_fn is not None:
            logger.warning("loss_fn in Pairwise model will be ignored")

        self.cross_encoder = cross_encoder

    def forward(
        self,
        inputs: Union[Dict[str, torch.Tensor], list],
        inputs_pair: Optional[Dict[str, torch.Tensor]] = None,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        if isinstance(inputs, (list, tuple)) and 2 <= len(inputs) <= 3 or inputs_pair is not None:
            if inputs_pair:
                input1 = inputs
                input2 = inputs_pair
            else:
                input1 = inputs[0]
                input2 = inputs[1]
                if len(inputs) == 3:
                    input3 = inputs[2]

            if self.cross_encoder:
                ids1, mask1 = input1['input_ids'], input1['attention_mask']
                ids2, mask2 = input2['input_ids'], input2['attention_mask']
                ids = torch.cat([ids1, ids2], dim=0)
                mask = torch.cat([mask1, mask2], dim=0)

                transformer_out = super().forward_from_loader(
                    {"input_ids": ids, "attention_mask": mask}, without_pooling=True
                )
                pooled_output = self.pooling(transformer_out[0], mask)
                pooled_output1 = pooled_output[: len(ids1), :]
                pooled_output2 = pooled_output[len(ids1) :, :]
                return pooled_output1, pooled_output2
            else:
                # bi-encoder, pooling in each
                pooled_output1 = super().forward(input1)
                pooled_output2 = super().forward(input2)
                if len(inputs) == 3:
                    pooled_output3 = super().forward(input3)
                    return pooled_output1, pooled_output2, pooled_output3
                return pooled_output1, pooled_output2
        else:
            pooled_output = super(PairwiseModel, self).forward_from_loader(inputs, without_pooling=False)
            return pooled_output


def unsorted_segment_mean(data: torch.Tensor, segment_ids: torch.Tensor, num_segments: int) -> torch.Tensor:
    result_shape = (num_segments, data.size(1))
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result = data.new_full(result_shape, 0)  # init empty result tensor
    count = data.new_full(result_shape, 0)
    result.scatter_add_(0, segment_ids, data)
    count.scatter_add_(0, segment_ids, torch.ones_like(data))
    return result / count.clamp(min=1)


class ListwiseModel(AutoModelForEmbedding):
    """
    segment_id
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        loss_fn: Optional[Callable] = None,
        listwise_pooling: bool = False,
        num_segments: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            normalize_embeddings=normalize_embeddings,
            loss_fn=loss_fn,
            **kwargs,
        )
        self.pooling_method = pooling_method
        self.num_segments = num_segments

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        encoding = super().forward(inputs)

        res = dict()
        if self.pooling_method == 'unsorted_segment_mean':
            encoding = unsorted_segment_mean(
                encoding, segment_ids=inputs['segment_ids'], num_segments=self.num_segments
            )
            res['pred'] = self.fc(encoding[:, 1:])
        else:
            encodings = []
            for i in range(self.num_segments):
                mask_ = (inputs['segment_ids'] == i + 1).int()
                encoding_ = self.pooling(encoding, mask_)
                encoding_ = self.fc(encoding_)
                encodings.append(encoding_)
            res['pred'] = torch.stack(encodings, 1)

        res['pred'] = res['pred'].squeeze(-1)
        return res



from .collator import ColBertCollator, PairCollator, RerankCollator, TripletCollator
from .dataset import RerankDataset, RetrievalDataset
from .sampler import GroupedBatchSampler, GroupSortedBatchSampler



import re
from typing import Dict, List, Literal

import torch

DEFAULT_LLM_PATTERNS = [r'.*llama.*', r'.*qwen.*', r'.*baichuan.*', r'.*mistral.*', r'.*intern.*']


def get_device_name() -> Literal["mps", "cuda", "cpu"]:
    """
    Returns the name of the device where this module is running on.
    It's a simple implementation that doesn't cover cases when more powerful GPUs are available and
    not a primary device ('cuda:0') or MPS device is available, but not configured properly:
    https://pytorch.org/docs/master/notes/mps.html

    :return: Device name, like 'cuda' or 'cpu'
    """
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"


def batch_to_device(batch: Dict, target_device: str) -> Dict[str, torch.Tensor]:
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], torch.Tensor):
            batch[key] = batch[key].to(target_device)
        else:
            batch[key] = torch.tensor(batch[key], dtype=torch.long).to(target_device)
    return batch


def check_casual_lm(model_name_or_path: str, llm_regex_patterns: List[str] = None) -> bool:
    if llm_regex_patterns is not None:
        llm_regex_patterns += DEFAULT_LLM_PATTERNS
    else:
        llm_regex_patterns = DEFAULT_LLM_PATTERNS
    model_name_or_path = model_name_or_path.lower()
    for pattern in llm_regex_patterns:
        if re.match(pattern, model_name_or_path):
            return True
    return False



"""
refer and rights reserved by:
https://github.com/InternLM/HuixiangDou
"""

import hashlib
import io
import logging

import pandas as pd
import requests

logger = logging.getLogger(__name__)


class FileParser:
    def __init__(self):
        self.image_suffix = ['.jpg', '.jpeg', '.png', '.bmp']
        self.md_suffix = ['.md']
        self.text_suffix = ['.txt', '.text']
        self.excel_suffix = ['.xlsx', '.xls', '.csv']
        self.pdf_suffix = ['.pdf']
        self.ppt_suffix = ['.pptx']
        self.html_suffix = ['.html', '.htm', '.shtml', '.xhtml']
        self.word_suffix = ['.docx', '.doc']
        self.normal_suffix = (
            self.md_suffix
            + self.text_suffix
            + self.excel_suffix
            + self.pdf_suffix
            + self.word_suffix
            + self.ppt_suffix
            + self.html_suffix
        )

    def read(self, filepath: str, to_document: bool = False):
        file_type = self.get_type(filepath)
        text = ''

        try:
            if file_type == 'md' or file_type == 'text':
                with open(filepath) as f:
                    text = f.read()

            elif file_type == 'pdf':
                text += self.read_pdf(filepath)

            elif file_type == 'excel':
                text += self.read_excel(filepath)

            elif file_type == 'word' or file_type == 'ppt':
                import textract

                text = textract.process(filepath).decode('utf8')
                if file_type == 'ppt':
                    text = text.replace('\n', ' ')

            elif file_type == 'html':
                from bs4 import BeautifulSoup

                with open(filepath) as f:
                    soup = BeautifulSoup(f.read(), 'html.parser')
                    text += soup.text

        except Exception as e:
            logger.error((filepath, str(e)))
            return '', e

        text = text.replace('\n\n\n', '\n')
        text = text.replace('\n\n', '\n')
        text = text.replace('  ', ' ')
        return text, None

    def get_type(self, filepath: str):
        for suffix in self.pdf_suffix:
            if filepath.endswith(suffix):
                return 'pdf'

        for suffix in self.md_suffix:
            if filepath.endswith(suffix):
                return 'md'

        for suffix in self.ppt_suffix:
            if filepath.endswith(suffix):
                return 'ppt'

        for suffix in self.image_suffix:
            if filepath.endswith(suffix):
                return 'image'

        for suffix in self.text_suffix:
            if filepath.endswith(suffix):
                return 'text'

        for suffix in self.word_suffix:
            if filepath.endswith(suffix):
                return 'word'

        for suffix in self.excel_suffix:
            if filepath.endswith(suffix):
                return 'excel'

        for suffix in self.html_suffix:
            if filepath.endswith(suffix):
                return 'html'
        return None

    def md5(self, filepath: str):
        hash_object = hashlib.sha256()
        with open(filepath, 'rb') as file:
            chunk_size = 8192
            while chunk := file.read(chunk_size):
                hash_object.update(chunk)

        return hash_object.hexdigest()[0:8]

    def read_pdf(self, filename):
        with open(filename, 'rb') as file:
            text = self.extract_text_from_pdf(file)
        return text

    @staticmethod
    def read_txt(filename):
        with open(filename, 'r', encoding='utf-8') as file:
            return file.read()

    def read_url(self, url):
        response = requests.get(url)
        if url.endswith('.pdf'):
            return self.extract_text_from_pdf(io.BytesIO(response.content))
        elif url.endswith('.txt'):
            return response.text
        else:
            return "Unsupported file format"

    def read_excel(self, filepath: str):
        table = None
        if filepath.endswith('.csv'):
            table = pd.read_csv(filepath)
        else:
            table = pd.read_excel(filepath)
        if table is None:
            return ''
        json_text = table.dropna(axis=1).to_json(force_ascii=False)
        return json_text

    def read_epub(self, filepath: str):
        import ebooklib
        from bs4 import BeautifulSoup

        book = ebooklib.epub.read_epub(filepath)

        text_content = []
        for item in book.get_items():
            if item.get_type() == ebooklib.ITEM_DOCUMENT:
                text_content.append(item.get_content().decode("utf-8"))

        wt = "\n".join(text_content)
        sp = BeautifulSoup(wt, "html.parser")
        text = sp.get_text()
        return text

    @staticmethod
    def extract_text_from_pdf(file_stream):
        import pypdf

        text = ""
        reader = pypdf.PdfReader(file_stream)
        for page in reader.pages:
            text += page.extract_text()
        return text



import torch
import torch.nn as nn


class Dice(nn.Module):
    def __init__(self, feature_num):
        super(Dice, self).__init__()
        self.alpha = nn.Parameter(torch.ones(1))
        self.feature_num = feature_num
        self.bn = nn.BatchNorm1d(self.feature_num, affine=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x_norm = self.bn(x)
        x_p = self.sigmoid(x_norm)
        return self.alpha * (1.0 - x_p) * x + x_p * x



import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from transformers import AutoModel

from ..tools.file_parser import FileParser

logger = logging.getLogger(__name__)


class ModelCenter(object):
    def __init__(self):
        pass

    def chat(self, query: str, chat_history):
        return


class KnowledgeCenter(object):
    def __init__(self):
        self.parser = FileParser()

    def init_vector_db(self):
        pass

    def add_document(self, file_path: str):
        doc = self.parser.read(file_path)
        print(doc)


class RagPipe(object):
    def __init__(self):
        pass

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: Union[str, Path],
        n_gpu: int = -1,
        verbose: int = 1,
        index_root: Optional[str] = None,
    ):
        instance = cls()
        instance.model = AutoModel()
        return instance

    @classmethod
    def from_index(cls, index_path: Union[str, Path], n_gpu: int = -1, verbose: int = 1):
        instance = cls()
        index_path = Path(index_path)
        instance.model = AutoModel()

        return instance

    def add_to_index(self):
        return

    def encode(self):
        return

    def index(self):
        return

    def search(self):
        return


class Generator(object):
    def __init__(self, config_path: str):
        self.config_path = config_path

    def _load_config(self):
        # with open(self.config_path, encoding='utf8') as f:
        #     config = pytoml.load(f)
        #     return config['llm']
        pass

    def generate(self, prompt, history=None, remote=False):
        pass



import copy
import logging
import os
import random
from typing import Dict, Iterable, List, Optional, Tuple, Union

import datasets
from torch.utils.data import Dataset
from transformers import BatchEncoding, PreTrainedTokenizer

logger = logging.getLogger(__name__)


class RetrievalDataset(Dataset):
    def __init__(self, data_name_or_path: str, cache_dir: Optional[str] = None):
        if os.path.isdir(data_name_or_path):
            train_datasets = []
            for file in os.listdir(data_name_or_path):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(data_name_or_path, file),
                    split="train",
                    cache_dir=cache_dir,
                )

                train_datasets.append(temp_dataset)
            self.dataset = datasets.concatenate_datasets(train_datasets)
        else:
            self.dataset = datasets.load_dataset("json", data_files=data_name_or_path)

        if 'train' in self.dataset:
            self.dataset = self.dataset['train']

        # self.tokenizer = tokenizer
        logger.info("Loaded {} retrieval data.".format(len(self.dataset)))

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, item: int) -> Union[Dict[str, str], List[BatchEncoding]]:
        query = self.dataset[item]["query"]
        if isinstance(self.dataset[item]["pos"], Iterable):
            pos = random.choice(self.dataset[item]["pos"])
        else:
            pos = self.dataset[item]["pos"]
        if isinstance(self.dataset[item]["neg"], Iterable):
            neg = random.choice(self.dataset[item]["neg"])
        else:
            neg = self.dataset[item]["neg"][0]
        sample = {"query": query, "pos": pos, "neg": neg}
        return sample

    def dynamic_sample(self, batch_size: int, missing_list=None, wrong_dict=None, max_wrong: int = 16):
        logger.info('\nDynamic Shuffle Sample...')

        return

    @classmethod
    def from_local(cls):
        return


class RerankDataset(Dataset):
    def __init__(
        self,
        data_name_or_path: str,
        cache_dir: Optional[str] = None,
        query_key='query',
        positive_key: Optional[str] = None,
        negative_key: Optional[str] = None,
        negative_numbers: Optional[int] = None,
    ):
        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key
        self.negative_numbers = negative_numbers

        if os.path.isdir(data_name_or_path):
            train_datasets = []
            for file in os.listdir(data_name_or_path):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(data_name_or_path, file),
                    cache_dir=cache_dir,
                )

                train_datasets.append(temp_dataset)
            dataset = datasets.concatenate_datasets(train_datasets)

        else:
            dataset = datasets.load_dataset("json", data_files=data_name_or_path)

        if 'train' in dataset:
            dataset = dataset['train']

        if positive_key:
            dataset = self.generate_samples(dataset)

        self.dataset = dataset
        logger.info("Loaded {} rerank data.".format(len(self.dataset)))

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item: int):
        if isinstance(self.dataset[item], dict):
            query = self.dataset[item][self.query_key]
            document = self.dataset[item]['document']
            labels = self.dataset[item]['labels']
        else:
            query, document, labels = self.dataset[item]
        sample = {"query": query, "document": document, "labels": labels}
        return sample

    def generate_samples(self, dataset):
        samples: List = []
        for data in dataset:
            for text_pos in data[self.positive_key]:
                samples.append([data[self.query_key], text_pos, 1])

            negative_samples = data[self.negative_key]
            if self.negative_numbers:
                # TODO: random strategy
                negative_samples = negative_samples[: self.negative_numbers]
            for text_neg in negative_samples:
                samples.append([data[self.query_key], text_neg, 0])
        return samples

    @classmethod
    def from_local(cls):
        return



import gc
import logging
import math
import time
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch

# from accelerate import Accelerator
# from accelerate.logging import get_logger
# from accelerate.utils import set_seed
from torch import nn
from tqdm import tqdm

from .adversarial import AWP, EMA, FGM

logger = logging.getLogger(__name__)


def train_fn(
    epoch: int,
    model: nn.Module,
    train_loader,
    optimizer,
    criterion: Optional[Callable] = None,
    apex: bool = False,
    gradient_accumulation_steps: int = 1,
    max_grad_norm: float = 1,
    data_collator=None,
    batch_scheduler=None,
    fgm=None,
    awp=None,
    ema_inst=None,
    print_freq: int = 100,
    wandb: bool = False,
    device: str = "cuda",
    teacher: Optional[nn.Module] = None,
    teacher_loss: Optional[Callable] = None,
    **kwargs,
):
    start = time.time()
    global_step = 0
    model.train()
    scaler = torch.cuda.amp.GradScaler(enabled=apex)
    losses = AverageMeter()
    save_step = int(len(train_loader) / 1)

    # loss_list = []
    # metrics_list = []

    for step, inputs in enumerate(train_loader):
        if isinstance(inputs, (list, tuple)) and len(inputs) == 2:
            inputs, labels = inputs
        elif isinstance(inputs, dict):
            assert data_collator is not None, "if labels are in inputs keys, can't set data_collator to None"
            labels = inputs['labels']
            inputs.pop('labels', None)

        if data_collator:
            inputs = data_collator(inputs)

        if isinstance(inputs, (list, tuple)):
            for input in inputs:
                if isinstance(input, dict):
                    for k, v in input.items():
                        input[k] = v.to(device)
                else:
                    input = input.to(device)
        else:
            for k, v in inputs.items():
                inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            if criterion is None:
                try:
                    preds = model(inputs=inputs, labels=labels)
                except ValueError:
                    preds = model(**inputs, labels=labels)
            else:
                try:
                    preds = model(inputs=inputs)
                except ValueError:
                    preds = model(**inputs)
            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
            else:
                if isinstance(inputs, dict):
                    if isinstance(preds, dict):
                        logits = preds['logits']
                    elif isinstance(preds, (list, tuple)):
                        logits = preds[0]
                    else:
                        logits = preds

                    if 'weights' in inputs:
                        loss = criterion(logits, labels.float(), inputs['weights'])
                    else:
                        loss = criterion(logits, labels.float())
                else:
                    loss = criterion(preds[0], preds[1])

        if teacher:
            with torch.no_grad():
                teacher_output = teacher(inputs)
                loss_distill = teacher_loss(preds, teacher_output)
                loss = loss_distill

        if gradient_accumulation_steps > 1:
            loss = loss / gradient_accumulation_steps

        losses.update(loss.item(), batch_size)
        scaler.scale(loss).backward()

        # if isinstance(preds, dict):
        #     acc = (preds["feature"].softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # else:
        #     acc = (preds.softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # acc = acc.detach().cpu().item()
        # acc_list.append(acc)

        # ---------------------fgm-------------
        if fgm:
            # attach the embedding
            fgm.attack(epsilon=1.0)
            with torch.cuda.amp.autocast(enabled=apex):
                preds = model(inputs)
                if isinstance(preds, dict) and "loss" in preds:
                    loss_avg = preds["loss"]
                else:
                    loss_avg = criterion(preds, labels)
            if gradient_accumulation_steps > 1:
                loss_avg = loss_avg / gradient_accumulation_steps
            losses.update(loss_avg.item(), batch_size)
            scaler.scale(loss_avg).backward()
            # restore the embedding
            fgm.restore()

        grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)

        if (step + 1) % gradient_accumulation_steps == 0:
            scaler.step(optimizer)
            scaler.update()

            if ema_inst:
                ema_inst.update()

            optimizer.zero_grad()
            global_step += 1
            if batch_scheduler:
                batch_scheduler.step()

        if step % print_freq == 0 or (step + 1) == len(train_loader):
            print(
                "Epoch: [{0}][{1}/{2}] "
                "Elapsed {remain:s} "
                "Loss: {loss.val:.4f}({loss.avg:.4f}) "
                "Grad: {grad_norm:.4f}  "
                "LR: {lr:.8f}  "
                # "ACC: {acc: .4f}"
                .format(
                    epoch + 1,
                    step,
                    len(train_loader) - 1,
                    remain=timeSince(start, float(step + 1) / len(train_loader)),
                    loss=losses,
                    grad_norm=grad_norm,
                    lr=batch_scheduler.get_lr()[0] if batch_scheduler else optimizer.param_groups[0]["lr"],
                    # acc=np.mean(acc_list),
                )
            )

        if wandb and step % print_freq == 0:
            print({"[loss": losses.val, "[lr": optimizer.param_groups[0]["lr"]})

        if (step + 1) % save_step == 0 and epoch > -1:
            if ema_inst:
                ema_inst.apply_shadow()

            if ema_inst:
                ema_inst.restore()

    return losses.avg


@torch.no_grad()
def valid_fn(
    epoch,
    valid_loader,
    model,
    criterion,
    eval_metrics=None,
    apex: bool = False,
    device="cuda",
    data_collator=None,
    print_freq: int = 100,
):
    start = time.time()
    model.eval()
    losses = AverageMeter()
    pred_array = torch.tensor([], device=device)
    label_array = np.array([])

    for step, inputs in enumerate(valid_loader):
        if isinstance(inputs, (list, tuple)) and len(inputs) == 2:
            inputs, labels = inputs
        elif isinstance(inputs, dict):
            labels = inputs['labels']
            inputs.pop('labels', None)

        if data_collator:
            inputs = data_collator(inputs)

        for k, v in inputs.items():
            inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            try:
                preds = model(inputs, labels=labels)
            except ValueError:
                preds = model(inputs)
            else:
                logger.warning('inputs of model should be with or without labels')

            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
                pred = preds["output"]
            else:
                loss = criterion(preds, labels)
                pred = preds

        losses.update(loss.item(), batch_size)
        pred_array = torch.cat((pred_array, pred), dim=0)
        label_array = np.append(label_array, np.array(labels))

        if step % print_freq == 0 or step == (len(valid_loader) - 1):
            print(
                "Epoch: [{0}][{1}/{2}] "
                "Elapsed {remain:s} "
                "Loss: {loss.val:.4f}({loss.avg:.4f}) "
                # "ACC: {acc: .4f}"
                .format(
                    epoch + 1,
                    step,
                    len(valid_loader),
                    remain=timeSince(start, float(step + 1) / len(valid_loader)),
                    loss=losses,
                    # acc=np.mean(acc_list),
                )
            )

    if eval_metrics is not None:
        score = eval_metrics(label_array, pred_array)
        print(f"Epoch {epoch}, Score: {score}")
        return losses.avg, score

    return losses.avg


def inference_fn(test_loader, model, device):
    preds: List[np.ndarray] = []
    model.eval()
    model.to(device)
    tk0 = tqdm(test_loader, total=len(test_loader))
    for inputs in tk0:
        for k, v in inputs.items():
            inputs[k] = v.to(device)
        with torch.no_grad():
            y_preds = model(inputs)
        preds.append(y_preds.sigmoid().squeeze().to("cpu").numpy().reshape(-1))
    predictions = np.concatenate(preds)
    return predictions


class CustomTrainer(object):
    def __init__(self, model: Union[nn.Module], device: Optional[str] = None, apex: bool = False, teacher=None):
        if not device:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device
        self.model = model
        self.apex = apex
        self.train_fn = train_fn
        self.valid_fn = valid_fn
        if teacher:
            self.teacher = teacher.eval()
            self.teacher_loss = torch.nn.MSELoss()

    def train(
        self,
        train_loader,
        optimizer,
        epochs=10,
        criterion=None,
        scheduler=None,
        valid_loader=None,
        eval_metrics=None,
        data_collator=None,
        max_grad_norm=10,
        use_fgm: bool = False,
        use_awp: bool = False,
        ema_decay: float = 0,
        dynamic_margin_fn: Optional[Callable] = None,
        gradient_checkpointing: bool = False,
        resume_from_checkpoint: Optional[Union[str, bool]] = None,
        **kwargs,
    ):
        logger.info('-------START TO TRAIN-------')
        if resume_from_checkpoint is not None:
            checkpoint = torch.load(resume_from_checkpoint)
            self.model.load_state_dict(checkpoint)

        if use_fgm:
            logger.info('[FGM] Use FGM adversarial')
            fgm = FGM(self.model.to(self.device))
        elif use_awp:
            logger.info('[AWP] Use AWP adversarial')
            awp = AWP(self.model.to(self.device), optimizer)

        if ema_decay > 0:
            logger.info('[EMA] Use EMA while training')
            ema_inst = EMA(self.model.to(self.device), ema_decay)
            ema_inst.register()

        if gradient_checkpointing:
            self.model.gradient_checkpointing_enable()

        for epoch in range(epochs):
            if dynamic_margin_fn:
                margin = dynamic_margin_fn(epoch)
                logger.info(f"[Dynamic margin] Epoch: {epoch}, Margin: {margin}")
                if criterion:
                    criterion.set_margin(margin)
                elif self.model.loss_fn is not None:
                    self.model.loss_fn.set_margin(margin)
                else:
                    raise ValueError(
                        'dynamic margin is True, but criterion of trainer and loss_fn of model are both none'
                    )

            # lr = scheduler(optimizer, epoch)

            # 放在这里, 是因为前面的set_margin也会产生新的参数位于cpu
            # 如果把参数直接放在cuda, 则导致parameter里没有出现该参数，无法分别设置学习率
            self.model = self.model.to(self.device)
            optimizer.zero_grad()
            self.train_fn(
                epoch=epoch,
                train_loader=train_loader,
                model=self.model,
                criterion=criterion,
                optimizer=optimizer,
                batch_scheduler=scheduler,
                fgm=fgm if use_fgm else None,
                awp=awp if use_awp else None,
                ema_inst=ema_inst if ema_decay > 0 else None,
                apex=self.apex,
                gradient_accumulation_steps=1,
                max_grad_norm=max_grad_norm,
                valid_loader=valid_loader,
                data_collator=data_collator,
                device=self.device,
            )

            if valid_loader is not None:
                self.valid_fn(
                    epoch=epoch,
                    valid_loader=valid_loader,
                    model=self.model,
                    criterion=criterion,
                    eval_metrics=eval_metrics,
                    apex=self.apex,
                    device=self.device,
                    data_collator=None,
                )

        torch.cuda.empty_cache()
        gc.collect()

    def predict(self, test_loader):
        self.model.eval()
        tbar = tqdm(test_loader)

        vectors = []
        with torch.no_grad():
            for idx, (inputs, labels) in enumerate(tbar):
                for k, v in inputs.items():
                    inputs[k] = v.to(self.device)

                vec = self.model(inputs)
                if isinstance(vec, dict):
                    vec = vec["embeddings"]
                vectors.append(vec.detach().cpu().numpy())

        V = np.concatenate(vectors)
        return V

    def save_state(self, output_dir: Optional[str] = None):
        # TODO: remove the loss_fn while save weights
        if not output_dir:
            output_dir = "./model.pth"
        torch.save(self.model.state_dict(), output_dir)

    def save_model(self, output_dir: Optional[str] = None, state_dict=None):
        pass


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def asMinutes(s):
    m = math.floor(s / 60)
    s -= m * 60
    return "%dm %ds" % (m, s)


def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return "%s (remain %s)" % (asMinutes(s), asMinutes(rs))



import logging
import time
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from sklearn.neighbors import NearestNeighbors
from tqdm import tqdm
from tqdm.autonotebook import trange

logger = logging.getLogger(__name__)


class AutoModelForRetrieval(object):
    def __init__(
        self, embedding_model: Optional[nn.Module] = None, method: Literal['cosine', 'knn'] = "cosine"
    ) -> None:
        super().__init__()
        self.embedding_model = embedding_model
        self.method = method

    def similarity_search(
        self,
        query_embed: Union[torch.Tensor, np.ndarray],
        document_embed: Optional[torch.Tensor] = None,
        index_path: Optional[str] = None,
        top_k: int = 3,
        batch_size: int = -1,
        convert_to_numpy: bool = True,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if len(query_embed.shape) == 1:
            if isinstance(query_embed, np.ndarray):
                query_embed = query_embed.reshape(1, -1)
            else:
                query_embed = query_embed.view(1, -1)

        self.top_k = top_k
        if document_embed is None and index_path is None:
            logging.warning('Please provide either document_embed for knn/tensor search or index_path for faiss search')
            return

        if index_path is not None:
            import faiss

            start_time = time.time()
            faiss_index = faiss.read_index(index_path)
            logger.info(f'Loading faiss index successfully, elapsed time: {time.time()-start_time:.2}s')

            if batch_size < 1:
                dists, indices = faiss_index.search(query_embed.astype(np.float32), k=top_k)
            else:
                dists, indices = faiss_search(
                    query_embed=query_embed,
                    faiss_index=faiss_index,
                    top_k=top_k,
                    batch_size=batch_size,
                )
            return dists, indices

        elif self.method == "knn":
            neighbors_model = NearestNeighbors(n_neighbors=top_k, metric="cosine", n_jobs=-1)
            neighbors_model.fit(document_embed)
            dists, indices = neighbors_model.kneighbors(query_embed)

        elif self.method == "cosine":
            dists, indices = cosine_similarity_search(
                query_embed, document_embed, top_k=top_k, batch_size=batch_size, convert_to_numpy=convert_to_numpy
            )

        else:
            raise ValueError(
                f"Only 'cosine' and 'knn' method are supported by similarity_search, while get {self.method}"
            )

        return dists, indices

    def search(self):
        return

    def similarity(self, queries: Union[str, List[str]], keys: Union[str, List[str], np.ndarray]):
        return

    def get_relevant_documents(self, query: str):
        return

    def get_pandas_candidate(
        self,
        query_ids: Union[pd.Series, np.ndarray],
        document_ids: Union[pd.Series, np.ndarray],
        dists: np.ndarray,
        indices: np.ndarray,
    ) -> pd.DataFrame:
        if isinstance(query_ids, pd.Series):
            query_ids = query_ids.values
        if isinstance(document_ids, pd.Series):
            document_ids = document_ids.values

        retrieval = {
            'query_id': np.repeat(query_ids, self.top_k),
            'predict_id': document_ids[indices.ravel()],
            'score': dists.ravel(),
        }
        return pd.DataFrame(retrieval)

    def get_rerank_df(
        self,
        input_df: pd.DataFrame,
        query_key: str = 'query_id',
        document_key: str = 'document_id',
        predict_key: str = 'predict_id',
    ) -> pd.DataFrame:
        """
        1: the candidate is in ground truth pool
        2: the candidate is related or not with query
        """
        logger.info('Generate rerank samples based on the retrieval prediction with: query_id, document_ids, pred_ids')

        samples = []
        for _, row in tqdm(input_df.iterrows(), total=len(input_df)):
            query_id = row[query_key]
            document_ids = row[document_key]
            if pd.isna(row[predict_key]):
                print(f' Data error, no {predict_key} in input_df')
                continue
            else:
                predict_ids_list = row[predict_key].split()

            if pd.isna(document_ids):
                print(f' Data Error, the ground truth {document_key} is none')
                for id in predict_ids_list:
                    samples.append({query_key: query_id, document_key: id, 'label': 0})
            else:
                documents = document_ids.split()
                for id in predict_ids_list:
                    if id not in documents:
                        samples.append({query_key: query_id, document_key: id, 'label': 0})
                    else:
                        samples.append({query_key: query_id, document_key: id, 'label': 1})

        return pd.DataFrame(samples)


class EnsembleRetriever(object):
    def __init__(self, retrievers, weights=None):
        pass


def cosine_similarity_search(
    query_embed: torch.Tensor,
    document_embed: torch.Tensor,
    top_k: int = 1,
    batch_size: int = 128,
    penalty: bool = True,
    temperature: float = 0,
    convert_to_numpy: bool = True,
    show_progress_bar: bool = None,
):
    if len(query_embed.size()) == 1:
        query_embed = query_embed.view(1, -1)
    assert query_embed.size()[1] == document_embed.size()[1], (
        f"The embed Shape of query_embed and document_embed should be same, "
        f"while received query {query_embed.size()} and document {document_embed.size()}"
    )
    chunk = batch_size if batch_size > 0 else len(query_embed)
    embeddings_chunks = query_embed.split(chunk)

    dists = []
    indices = []
    for idx in trange(0, len(embeddings_chunks), desc="Batches", disable=not show_progress_bar):
        cos_sim_chunk = torch.matmul(embeddings_chunks[idx], document_embed.transpose(0, 1))
        cos_sim_chunk = torch.nan_to_num(cos_sim_chunk, nan=0.0)
        # if penalty:
        #     pen = ((contents["old_source_count"].values==0) & (contents["old_nonsource_count"].values==1))
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1 + pen*0.5).reshape(1,-1)).to(DEVICE)
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1-(contents["old_count"].values==0)*0.1).
        #     reshape(1,-1)).to(DEVICE)
        if temperature:
            cos_sim_chunk = cos_sim_chunk / temperature
        top_k = min(top_k, cos_sim_chunk.size(1))
        dists_chunk, indices_chunk = torch.topk(cos_sim_chunk, k=top_k, dim=1)
        dists.append(dists_chunk[:, :].detach().cpu())
        indices.append(indices_chunk[:, :].detach().cpu())
    dists = torch.cat(dists)
    indices = torch.cat(indices)
    if convert_to_numpy:
        dists = dists.numpy()
        indices = indices.numpy()
    return dists, indices


def faiss_search(
    query_embed: torch.Tensor,
    faiss_index,
    top_k: int = 100,
    batch_size: int = 128,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    1. Encode queries into dense embeddings;
    2. Search through faiss index
    """
    # query_embeddings = model.encode_queries(
    #     queries["query"], batch_size=batch_size, max_length=max_length
    # )
    query_size = len(query_embed)
    assert query_size > 0, 'Please make sure the query_embeddings is not empty'

    all_scores = []
    all_indices = []

    for i in tqdm(range(0, query_size, batch_size), desc="Searching"):
        j = min(i + batch_size, query_size)
        query_embedding = query_embed[i:j]
        score, index = faiss_index.search(query_embedding.astype(np.float32), k=top_k)
        all_scores.append(score)
        all_indices.append(index)

    all_scores = np.concatenate(all_scores, axis=0)
    all_indices = np.concatenate(all_indices, axis=0)
    return all_scores, all_indices



"""
`SimCSE: Simple Contrastive Learning of Sentence Embeddings
<https://arxiv.org/abs/2104.08821>`_
"""

import logging
from typing import Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import nn

logger = logging.getLogger(__name__)


class SimCSE(nn.Module):
    def __init__(
        self, criterion: Union[nn.Module, Callable], temperature: float = 0.05, dynamic_temperature: bool = False
    ):
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature
        if dynamic_temperature:
            # TODO: dynamic_temperature
            self.temperature = nn.Parameter(torch.tensor(temperature))

    def forward(
        self,
        query_embeddings: torch.Tensor,
        pos_embeddings: torch.Tensor,
        neg_embeddings: Optional[torch.Tensor] = None,
    ):
        y_true = torch.arange(0, query_embeddings.size(0), device=query_embeddings.device)

        sim = F.cosine_similarity(query_embeddings.unsqueeze(1), pos_embeddings.unsqueeze(0), dim=2)
        # sim = sim - torch.eye(pos_embeddings.shape[0], device=query_embeddings.device) * 1e12

        sim = sim / self.temperature
        loss = self.criterion(sim, y_true)
        loss = torch.mean(loss)
        return loss



import logging
from typing import Dict, List, Tuple, Union


def mrr(
    qrels: Dict[str, Dict[str, int]], results: Dict[str, Dict[str, float]], k_values: List[int]
) -> Tuple[Dict[str, float]]:
    MRR = {}

    for k in k_values:
        MRR[f"MRR@{k}"] = 0.0

    k_max, top_hits = max(k_values), {}
    logging.info("\n")

    for query_id, doc_scores in results.items():
        top_hits[query_id] = sorted(doc_scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]

    for query_id in top_hits:
        query_relevant_docs = set([doc_id for doc_id in qrels[query_id] if qrels[query_id][doc_id] > 0])
        for k in k_values:
            for rank, hit in enumerate(top_hits[query_id][0:k]):
                if hit[0] in query_relevant_docs:
                    MRR[f"MRR@{k}"] += 1.0 / (rank + 1)
                    break

    for k in k_values:
        MRR[f"MRR@{k}"] = round(MRR[f"MRR@{k}"] / len(qrels), 5)
        logging.info("MRR@{}: {:.4f}".format(k, MRR[f"MRR@{k}"]))

    return MRR



import os
import re
import shutil
from pathlib import Path

from setuptools import Command, find_packages, setup

version_file = 'src/retrievals/version.py'


def get_version():
    with open(version_file, 'r') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


extras = {}

_deps = [
    "Pillow>=10.0.1",
    "accelerate>=0.21.0",
    "peft",
    "av>=9.2.0",  # Latest version of PyAV (10.0.0) has issues with audio stream.
    "beautifulsoup4",
    "codecarbon>=1.2.0",
    "cookiecutter>=1.7.3",
    "dataclasses",
    "datasets!=2.5.0",
    "decord==0.6.0",
    "deepspeed>=0.9.3",
    "diffusers",
    "dill<0.3.5",
    "evaluate>=0.2.0",
    "faiss-cpu",
    "fastapi",
    "filelock",
    "flax>=0.4.1,<=0.7.0",
    "fsspec<2023.10.0",
    "ftfy",
    "fugashi>=1.0",
    "GitPython<3.1.19",
    "hf-doc-builder>=0.3.0",
    "huggingface-hub>=0.19.3",
    "importlib_metadata",
    "ipadic>=1.0.0,<2.0",
    "isort>=5.5.4",
    "jax>=0.4.1,<=0.4.13",
    "jaxlib>=0.4.1,<=0.4.13",
    "kenlm",
    "transformers",
    "nltk",
    "natten>=0.14.6,<0.16.0",
    "numpy>=1.17",
    "onnxconverter-common",
    "onnxruntime-tools>=1.4.2",
    "onnxruntime>=1.4.0",
    "optuna",
    "optax>=0.0.8",
    "packaging>=20.0",
    "parameterized",
    "phonemizer",
    "protobuf",
    "psutil",
    "pyyaml>=5.1",
    "pydantic",
    "pytest>=7.2.0",
    "pytest-timeout",
    "pytest-xdist",
    "python>=3.8.0",
    "regex!=2019.12.17",
    "requests",
    "rhoknp>=1.1.0",
    "rjieba",
    "rouge-score!=0.0.7,!=0.0.8,!=0.1,!=0.1.1",
    "ruff==0.1.5",
    "sacrebleu>=1.4.12",
    "sacremoses",
    "safetensors>=0.4.1",
    "sagemaker>=2.31.0",
    "scikit-learn",
    "sentencepiece>=0.1.91,!=0.1.92",
    "sigopt",
    "starlette",
    "sudachipy>=0.6.6",
    "sudachidict_core>=20220729",
    "timeout-decorator",
    "timm",
    "tokenizers>=0.14",
    "torch",
    "pyctcdecode>=0.4.0",
    "tqdm>=4.27",
    "unidic>=1.0.2",
    "unidic_lite>=1.0.7",
    "uvicorn",
]


# this is a lookup table with items like:
deps = {b: a for a, b in (re.findall(r"^(([^!=<>~ ]+)(?:[!=<>~ ].*)?$)", x)[0] for x in _deps)}

install_requires = [
    deps["filelock"],  # filesystem locks, e.g., to prevent parallel downloads
    deps["huggingface-hub"],
    deps["numpy"],
    deps["packaging"],  # utilities from PyPA to e.g., compare versions
    deps["pyyaml"],  # used for the model cards metadata
    deps["regex"],  # for OpenAI GPT
    deps["requests"],  # for downloading models over HTTPS
    deps["tokenizers"],
    deps["safetensors"],
    deps["tqdm"],  # progress bars in model download and training scripts
]

setup(
    name="open-retrievals",
    version=get_version(),
    author="Longxing Tan",
    author_email="tanlongxing888@163.com",
    description="Text Embeddings for Retrieval and RAG based on transformers",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="NLP retrieval RAG rerank text embedding contrastive",
    license="Apache 2.0 License",
    url="https://github.com/LongxingTan/open-retrievals",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={"": ["**/*.cu", "**/*.cpp", "**/*.cuh", "**/*.h", "**/*.pyx"]},
    zip_safe=False,
    extras_require=extras,
    python_requires=">=3.8.0",
    install_requires=list(install_requires),
    classifiers=[
        # "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import logging
from copy import deepcopy
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from tqdm.auto import tqdm
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    PreTrainedTokenizer,
)
from transformers.modeling_outputs import (
    BaseModelOutputWithPooling,
    SequenceClassifierOutput,
)

from ..data.collator import RerankCollator
from .pooling import AutoPooling
from .utils import check_casual_lm, get_device_name

logger = logging.getLogger(__name__)


class RerankModel(nn.Module):
    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'mean',
        loss_fn: Union[nn.Module, Callable] = None,
        loss_type: Literal['classification', 'regression'] = 'classification',
        max_length: Optional[int] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()

        self.model: Optional[nn.Module] = model
        self.tokenizer = tokenizer
        self.pooling_method = pooling_method
        self.pooling = AutoPooling(self.pooling_method)

        if self.model:
            num_features: int = self.model.config.hidden_size
            self.classifier = nn.Linear(num_features, 1)
            self._init_weights(self.classifier)
        self.loss_fn = loss_fn
        self.loss_type = loss_type

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)

        self.max_length = max_length

        if device is None:
            self.device = get_device_name()
        else:
            self.device = device
        self.model.to(self.device)

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.model.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.model.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def encode(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        outputs: SequenceClassifierOutput = self.model(input_ids, attention_mask, output_hidden_states=True)
        if hasattr(outputs, 'last_hidden_state'):
            hidden_state = outputs.last_hidden_state
        else:
            hidden_state = outputs.hidden_states[1]

        if self.pooling:
            embeddings = self.pooling(hidden_state, attention_mask)
        else:
            embeddings = self.classifier(hidden_state[:, 1:])
        return embeddings

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        inputs: Optional[Dict[str, torch.Tensor]] = None,
        labels: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = True,
        **kwargs,
    ) -> Union[Dict[str, torch.Tensor], Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        if input_ids is not None:
            features = self.encode(input_ids=input_ids, attention_mask=attention_mask)
        elif inputs is not None:
            features = self.encode(**inputs)
        else:
            raise ValueError("input_ids(tensor) and inputs(dict) can't be empty as the same time")
        logits = self.classifier(features).reshape(-1)

        if return_dict:
            outputs_dict = dict()
            outputs_dict['logits'] = logits

        if labels is not None:
            if not self.loss_fn:
                if self.loss_type == 'regression':
                    logits = torch.sigmoid(logits)
                    self.loss_fn = nn.MSELoss()

                elif self.loss_type == 'classification':
                    self.loss_fn = nn.BCEWithLogitsLoss(reduction='mean')

            loss = self.loss_fn(logits, labels.float())
            if return_dict:
                outputs_dict['loss'] = loss
                return outputs_dict
            else:
                return logits, loss
        else:
            if return_dict:
                return outputs_dict
            return logits

    def set_model_type(self, model_type: Literal['cross-encoder', 'colbert'], **kwargs):
        model_class = {'cross-encoder': self, 'colbert': ColBERT}
        model_class = model_class.get(model_type.lower())
        return model_class(
            model=self.model,
            tokenizer=self.tokenizer,
            pooling_method=self.pooling_method,
            loss_fn=self.loss_fn,
            loss_type=self.loss_type,
            **kwargs,
        )

    @torch.no_grad()
    def compute_score(
        self,
        text_pairs: Union[List[Tuple[str, str]], Tuple[str, str]],
        data_collator: Optional[RerankCollator] = None,
        batch_size: int = 128,
        max_length: int = 512,
        normalize: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ):
        self.model.eval()
        if isinstance(text_pairs[0], str):
            text_pairs = [text_pairs]

        batch_size = min(batch_size, len(text_pairs))

        # if not data_collator:
        #     data_collator = RerankCollator(tokenizer=self.tokenizer)

        scores_list: List[float] = []
        for i in tqdm(range(0, len(text_pairs), batch_size), desc="Scoring", disable=not show_progress_bar):
            if isinstance(text_pairs[0][0], str):
                batch = self.tokenizer(
                    text_pairs[i : i + batch_size],
                    padding=True,
                    truncation=True,
                    max_length=max_length,
                    return_tensors="pt",
                )
            else:
                batch = self.tokenizer.pad(
                    text_pairs[i : i + batch_size],
                    padding=True,
                    max_length=None,
                    pad_to_multiple_of=None,
                    return_tensors='pt',
                )

            batch_on_device = {k: v.to(self.device) for k, v in batch.items()}
            scores = self.model(**batch_on_device, return_dict=True).logits.view(-1).float()
            if normalize:
                scores = torch.sigmoid(scores)
            scores_list.extend(scores.cpu().numpy().tolist())

        if len(scores_list) == 1:
            return scores_list[0]
        return scores_list

    @torch.no_grad()
    def rerank(
        self,
        query: str,
        documents: List[str],
        data_collator: Optional[RerankCollator] = None,
        batch_size: int = 32,
        chunk_max_length: int = 512,
        chunk_overlap: int = 50,
        max_chunks_per_doc: int = 100,
        normalize: bool = False,
        show_progress_bar: bool = None,
        return_dict: bool = True,
        **kwargs,
    ):
        # if isinstance(query, str):
        #     text_pairs = [(query, doc) for doc in document]
        # elif isinstance(query, (list, tuple)):
        #     text_pairs = [(q, doc) for q, doc in zip(query, document)]
        # else:
        #     pass

        if query is None or len(query) == 0 or len(documents) == 0:
            return {'rerank_documents': [], 'rerank_scores': []}

        splitter = DocumentSplitter(
            chunk_size=chunk_max_length, chunk_overlap=chunk_overlap, max_chunks_per_doc=max_chunks_per_doc
        )
        text_pairs, sentence_pairs_pids = splitter.create_documents(
            query,
            documents,
            tokenizer=self.tokenizer,
        )

        tot_scores = self.compute_score(
            text_pairs=text_pairs,
            data_collator=data_collator,
            batch_size=batch_size,
            normalize=normalize,
            show_progress_bar=show_progress_bar,
        )

        merge_scores = [0 for _ in range(len(documents))]
        for pid, score in zip(sentence_pairs_pids, tot_scores):
            merge_scores[pid] = max(merge_scores[pid], score)

        merge_scores_argsort = np.argsort(merge_scores)[::-1]
        sorted_document = []
        sorted_scores = []
        for mid in merge_scores_argsort:
            sorted_scores.append(merge_scores[mid])
            sorted_document.append(documents[mid])

        if return_dict:
            return {
                'rerank_document': sorted_document,
                'rerank_scores': sorted_scores,
                'rerank_ids': merge_scores_argsort.tolist(),
            }
        else:
            return sorted_document

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        pooling_method: str = 'mean',
        loss_fn: Union[nn.Module, Callable] = None,
        loss_type: Literal['classification', 'regression'] = 'classification',
        num_labels: int = 1,
        causal_lm: bool = False,
        gradient_checkpointing: bool = False,
        trust_remote_code: bool = True,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        device: Optional[str] = None,
        **kwargs,
    ):
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        model = AutoModelForSequenceClassification.from_pretrained(
            model_name_or_path, num_labels=num_labels, trust_remote_code=trust_remote_code, **kwargs
        )
        if gradient_checkpointing:
            model.graident_checkpointing_enable()

        if use_fp16:
            model.half()
        if use_lora:
            from peft import LoraConfig, TaskType, get_peft_model

            if not lora_config:
                raise ValueError("If use_lora is true, please provide a valid lora_config from peft")
            model = get_peft_model(model, lora_config)
            model.print_trainable_parameters()

        reranker = cls(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            device=device,
            loss_fn=loss_fn,
            loss_type=loss_type,
        )
        return reranker

    def save_pretrained(self, path: str, safe_serialization: bool = True):
        """
        Saves all model and tokenizer to path
        """
        logger.info("Save model to {}".format(path))
        state_dict = self.model.state_dict()
        state_dict = type(state_dict)({k: v.clone().cpu() for k, v in state_dict.items()})
        self.model.save_pretrained(path, state_dict=state_dict, safe_serialization=safe_serialization)
        self.tokenizer.save_pretrained(path)


class ColBERT(RerankModel):
    def __init__(
        self,
        colbert_dim: int = 1,
        **kwargs,
    ):
        self.colbert_dim = colbert_dim
        super(ColBERT, self).__init__(**kwargs)

    def forward(
        self,
        query_input_ids: Optional[torch.Tensor],
        query_attention_mask: Optional[torch.Tensor],
        pos_input_ids: Optional[torch.Tensor],
        pos_attention_mask: Optional[torch.Tensor],
        neg_input_ids: Optional[torch.Tensor] = None,
        neg_attention_mask: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = True,
        **kwargs,
    ):
        query_embedding = self.encode(query_input_ids, query_attention_mask)
        pos_embedding = self.encode(pos_input_ids, pos_attention_mask)

        if neg_input_ids and neg_attention_mask:
            neg_embedding = self.encode(neg_input_ids, neg_attention_mask)
        else:
            neg_embedding = None

        loss = self.loss_fn(query_embedding, pos_embedding, neg_embedding)
        if return_dict:
            outputs_dict = dict()
            outputs_dict['loss'] = loss
            return outputs_dict
        return loss

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        pooling_method: str = 'mean',
        loss_type: Literal['classification', 'regression'] = 'classification',
        num_labels: int = 1,
        causal_lm: bool = False,
        gradient_checkpointing: bool = False,
        trust_remote_code: bool = True,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        device: Optional[str] = None,
        colbert_dim: int = 1,
        **kwargs,
    ):
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        model = AutoModelForSequenceClassification.from_pretrained(
            model_name_or_path, num_labels=num_labels, trust_remote_code=trust_remote_code, **kwargs
        )
        if gradient_checkpointing:
            model.graident_checkpointing_enable()

        if use_fp16:
            model.half()
        if use_lora:
            # peft config and wrapping
            from peft import LoraConfig, TaskType, get_peft_model

            if not lora_config:
                raise ValueError("If use_lora is true, please provide a valid lora_config from peft")
            model = get_peft_model(model, lora_config)
            model.print_trainable_parameters()

        reranker = cls(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            device=device,
            loss_type=loss_type,
            colbert_dim=colbert_dim,
        )
        return reranker


class DocumentSplitter(object):
    """
    Rerank the long document
    - https://github.com/netease-youdao/BCEmbedding/blob/master/BCEmbedding/models/utils.py
    """

    def __init__(self, chunk_size: int, chunk_overlap: int = 0, max_chunks_per_doc: int = 32):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.max_chunks_per_doc = max_chunks_per_doc

    def create_documents(self, query, documents, tokenizer):
        res_merge_inputs = []
        res_merge_inputs_pids = []

        query_inputs = tokenizer.encode_plus(query, truncation=False, padding=False)
        sep_id = tokenizer.sep_token_id
        doc_max_length = self.chunk_size - len(query_inputs['input_ids']) - 2

        for pid, document in enumerate(documents):
            document_inputs = tokenizer.encode_plus(document, truncation=False, padding=False, add_special_tokens=False)
            doc_inputs_length = len(document_inputs['input_ids'])

            if doc_inputs_length <= doc_max_length:
                qc_merge_inputs = self._merge_inputs(query_inputs, document_inputs, sep_id)
                res_merge_inputs.append(qc_merge_inputs)
                res_merge_inputs_pids.append(pid)
            else:
                start_id = 0
                while start_id < doc_inputs_length:
                    end_id = start_id + doc_max_length
                    sub_document_inputs = {k: v[start_id:end_id] for k, v in document_inputs.items()}
                    start_id = end_id - self.chunk_overlap if end_id < doc_inputs_length else end_id

                    qp_merge_inputs = self._merge_inputs(query_inputs, sub_document_inputs, sep_id)
                    res_merge_inputs.append(qp_merge_inputs)
                    res_merge_inputs_pids.append(pid)
        return res_merge_inputs, res_merge_inputs_pids

    def _merge_inputs(self, chunk1_raw, chunk2, sep_id: int):
        chunk1 = deepcopy(chunk1_raw)

        chunk1['input_ids'].append(sep_id)
        chunk1['input_ids'].extend(chunk2['input_ids'])
        chunk1['input_ids'].append(sep_id)

        chunk1['attention_mask'].append(chunk2['attention_mask'][0])
        chunk1['attention_mask'].extend(chunk2['attention_mask'])
        chunk1['attention_mask'].append(chunk2['attention_mask'][0])

        if 'token_type_ids' in chunk1:
            token_type_ids = [1 for _ in range(len(chunk2['token_type_ids']) + 2)]
            chunk1['token_type_ids'].extend(token_type_ids)
        return chunk1
